; +
;    .
;
LdrImageNtHeader proc ImageBase:PVOID, ImageHeader:PIMAGE_NT_HEADERS
	mov edx,ImageBase
	mov eax,STATUS_INVALID_IMAGE_FORMAT
	assume edx:PIMAGE_DOS_HEADER
	cmp [edx].e_magic,'ZM'
	jne @f
	add edx,[edx].e_lfanew
	assume edx:PIMAGE_NT_HEADERS
	cmp [edx].Signature,'EP'
	jne @f
	cmp [edx].FileHeader.SizeOfOptionalHeader,sizeof(IMAGE_OPTIONAL_HEADER32)
	jne @f
	cmp [edx].FileHeader.Machine,IMAGE_FILE_MACHINE_I386	
	jne @f
	test [edx].FileHeader.Characteristics,IMAGE_FILE_32BIT_MACHINE
	je @f
	mov ecx,ImageHeader
	xor eax,eax
	mov dword ptr [ecx],edx
@@:
	ret
LdrImageNtHeader endp

; +
;
CompareAsciizString proc uses ebx String1:PSTR, String2:PSTR
    mov ecx,String1
    mov edx,String2
    xor ebx,ebx
@@:
    mov al,B[ecx + ebx]
    cmp B[edx + ebx],al
    jne @f
    inc ebx
    test al,al
    jne @b
@@:
    ret
CompareAsciizString endp

; +
;
LdrCalculateHash proc uses ebx esi PartialHash:ULONG, StrName:PCHAR, NameLength:ULONG
	xor eax,eax
	mov ecx,NameLength
	mov esi,StrName
	mov ebx,PartialHash
	cld
@@:
	lodsb
	xor ebx,eax
	xor ebx,ecx
	rol ebx,cl
	dec ecx
	jnz @b
	mov eax,ebx
	ret
LdrCalculateHash endp

; +
;    /  .
;
LdrImageQueryEntryFromHash proc uses ebx esi edi ImageBase:PVOID, HashOrFunctionName:DWORD, pComputeHashRoutine:PVOID, PartialHash:ULONG, Function:PVOID
Local ExportDirectory:PIMAGE_EXPORT_DIRECTORY
Local ImageHeader:PIMAGE_NT_HEADERS
Local NumberOfNames:ULONG
	mov ebx,ImageBase
	.if !Ebx
		assume fs:nothing
		mov eax,fs:[TEB.Peb]
		mov eax,PEB.Ldr[eax]
		mov eax,PEB_LDR_DATA.InLoadOrderModuleList.Flink[eax]
		mov eax,LDR_DATA_TABLE_ENTRY.InLoadOrderModuleList.Flink[eax]
		mov ebx,LDR_DATA_TABLE_ENTRY.DllBase[eax]	; ntdll.dll
	.endif
	invoke LdrImageNtHeader, Ebx, addr ImageHeader
	test eax,eax
	mov edx,ImageHeader
	jnz Exit
	assume edx:PIMAGE_NT_HEADERS
	mov eax,[edx].OptionalHeader.DataDirectory.VirtualAddress
	test eax,eax
	jz ErrImage
	add eax,ebx
	assume eax:PIMAGE_EXPORT_DIRECTORY
	mov ExportDirectory,eax
	mov esi,[eax].AddressOfNames
	test esi,esi
	jz ErrTable
	mov eax,[eax].NumberOfNames
	test eax,eax
	jz ErrTable
	mov NumberOfNames,eax
	add esi,ebx
	xor edi,edi
	cld
Next:
	mov eax,D[esi]
	add eax,ebx
	.if pComputeHashRoutine != NULL
		push edi
		mov edi,eax
		mov ecx,MAX_PATH
		mov edx,edi
		xor eax,eax
		repne scasb
		not ecx
		pop edi
		add ecx,MAX_PATH
		push ecx
		push edx
		push PartialHash
		Call pComputeHashRoutine
		cmp HashOrFunctionName,eax
	.else
		invoke CompareAsciizString, HashOrFunctionName, Eax
	.endif
	jnz @f
	mov ecx,ExportDirectory
	assume ecx:PIMAGE_EXPORT_DIRECTORY
	mov eax,[ecx].AddressOfNameOrdinals
	add eax,ebx
	movzx edi,W[2*edi+eax]
	.if edi
		.if edi >= [ecx]._Base
			sub edi,[ecx]._Base
  		.endif
		inc edi
	.endif
	mov esi,[ecx].AddressOfFunctions
	add esi,ebx
	mov ecx,D[4*edi + esi]
	test ecx,ecx
	mov edx,Function
	jz ErrImage
	add ecx,ebx
	xor eax,eax
	mov D[edx],ecx
	jmp Exit
@@:
	add esi,4
	inc edi
	dec NumberOfNames
	jnz Next
	mov eax,STATUS_PROCEDURE_NOT_FOUND
Exit:
	ret
ErrImage:
	mov eax,STATUS_INVALID_IMAGE_FORMAT
	jmp Exit
ErrTable:
	mov eax,STATUS_BAD_FUNCTION_TABLE
	jmp Exit
LdrImageQueryEntryFromHash endp

; +
;      .
;
LdrEncodeEntriesList proc uses esi edi ImageBase:PVOID, EntriesList:PVOID
	mov esi,EntriesList
	mov edi,esi
	lodsd
	.repeat
		invoke LdrImageQueryEntryFromHash, ImageBase, Eax, offset LdrCalculateHash, 0, Edi
		test eax,eax
		jnz Exit
		lodsd
		add edi,4
	.until !Eax
Exit:
	ret
LdrEncodeEntriesList endp

; +
;   .
;
; o R/W
;
LdrMapViewOfImage proc uses ebx Ip:PVOID, ViewBase:PPVOID
Local FileU[MAX_PATH*2 + sizeof(UNICODE_STRING)]:CHAR
Local ObjAttr:OBJECT_ATTRIBUTES
Local File:HANDLE, Section:HANDLE
Local IoStatus:IO_STATUS_BLOCK
Local ViewSize:ULONG
	lea eax,FileU
	mov ecx,Ip
	.if !Ecx
		mov ecx,fs:[TEB.Peb]
		mov ecx,PEB.LoaderLock[ecx]
	.endif
	push NULL
	push sizeof UNICODE_STRING  + sizeof MAX_PATH
	push eax
	push MemoryMappedFilenameInformation
	push ecx
	push NtCurrentProcess
	Call ZwQueryVirtualMemory
	%DBG "FLT: LdrMapViewOfImage.ZwQueryVirtualMemory(): 0x%X", Eax
	test eax,eax
	mov ObjAttr.uLength,sizeof(OBJECT_ATTRIBUTES)
	mov ObjAttr.uAttributes,OBJ_CASE_INSENSITIVE
	lea ecx,FileU
	jnz Exit	; STATUS_INVALID_ADDRESS/STATUS_INVALID_IMAGE_NOT_MZ/STATUS_FILE_INVALID etc.
	%DBG "FLT: LdrMapViewOfImage.FILE: %wZ", Ecx
	mov ObjAttr.hRootDirectory,eax
	mov ObjAttr.pObjectName,ecx
	lea edx,IoStatus
	push FILE_NON_DIRECTORY_FILE or FILE_SYNCHRONOUS_IO_NONALERT
	mov ObjAttr.pSecurityDescriptor,eax
	lea ecx,ObjAttr
	push FILE_SHARE_READ or FILE_SHARE_DELETE
	mov ObjAttr.pSecurityQualityOfService,eax
	push edx
	lea eax,File
	push ecx
	push SYNCHRONIZE or FILE_READ_ACCESS
	push eax
	Call ZwOpenFile
	%DBG "FLT: LdrMapViewOfImage.ZwOpenFile(): 0x%X", Eax
	test eax,eax
	lea ecx,Section
	jl Exit
	push File
	push SEC_IMAGE
	push PAGE_READONLY
	push NULL
	push NULL
	push SECTION_MAP_READ or SECTION_QUERY
	push ecx
	Call ZwCreateSection
	%DBG "FLT: LdrMapViewOfImage.ZwCreateSection(): 0x%X", Eax
	test eax,eax
	lea ecx,ViewSize
	.if Zero?
		mov ViewSize,eax
		push PAGE_READONLY
		push eax
		push ViewShare
		push ecx
		push eax
		push eax
		push eax
		push ViewBase
		push NtCurrentProcess
		push Section
		Call ZwMapViewOfSection
		%DBG "FLT: LdrMapViewOfImage.ZwMapViewOfSection(): 0x%X", Eax
		push eax	; STATUS_SUCCESS/STATUS_IMAGE_NOT_AT_BASE
		push Section
		Call ZwClose
		pop eax
	.endif
	push eax
	push File
	Call ZwClose
	pop eax
Exit:
	ret
LdrMapViewOfImage endp